local util = require "util"
local _M = {}
local mt = { __index = _M }
local cache = require "cache"
--The md5 of config string
_M.config_hash = nil

_M["configs"] = {}
_M['cache_rule'] = {}
-- _M["site_configs"] = {}

_M.cprefix = "/www/config/"

_M.configPath = ''
_M.default_html = nil
function _M:showdefault()

    if _M.default_html == nil then
        local path = _M.home_path() .."/configs/index.html"
        local f = io.open( path, 'r' )
        if f ~= nil then
            _M.default_html = f:read("*all")
            f:close()
        end
    end
    local html = util.string_replace( _M.default_html,'{{eventID}}', ngx.var.eventsid, 1 )
        
        ngx.header.content_type = "text/html"
        ngx.header['cache-control'] = "no-cache, no-store, must-revalidate"
        ngx.header['pragma'] = "no-cache"
        ngx.header['expires'] = "0"
        ngx.header.charset = "utf-8"
        ngx.say( html )
        
        ngx.exit(200)

end

local dkjson = require "dkjson"
local json = require "json"


function _M.new()
    local instance = {
        site_configs = nil
    }
    return setmetatable(instance, mt)
end

function _M:home_path()
    local current_script_path = debug.getinfo(1, "S").source:sub(2)
    local home_path = current_script_path:sub( 1, 0 - string.len("/script/CdnConfig.lua") -1 ) 
    return home_path
end

function _M:getHostConfig(domain, host, types)
    -- ngx.log(ngx.DEBUG, "domain host types".. domain..host..types)
    if domain == nil then
        domain = ""
    end
    if host == nil then
        host = ""
    end
    -- ngx.header['Content-Type'] = 'text/html'
    -- ngx.say(_M.configs["data"] .. "?type=" .. types .. "&domain=" .. domain .. "&host=" .. host)
    local ok, res = pcall(function()
        return util.http_get(_M.configs["data"] .. "?type=" .. types .. "&domain=" .. domain .. "&host=" .. host)
    end
    )
    if not ok then
        ngx.log(ngx.DEBUG, "empty http")
        res = nil
        return 
    end
    
    if res == 'no' then
        return res
    end
    -- 加到 domain_list
    if types == "domain" then
        local domain_list = ngx.shared.domain_list:get('all')
        if domain_list == nil then
            domain_list = ""
        end
        local arr = util.split(domain_list, ',')
        local find = false
        for _, v in pairs(arr) do
            if v == domain then 
                find = true
                break
            end
        end
        -- ngx.log(ngx.DEBUG, "hi domain_list: "..domain_list..','..#arr)
        if find == false then 
            table.insert(arr, domain)
            ngx.shared.domain_list:set('all', table.concat(arr, ','))
        end
        -- ngx.log()
    elseif types == "all" then
        -- 更新host的时候顺便更新ssl，所以不用多写
        local domain_list = ngx.shared.host_list:get('all')
        if domain_list == nil then
            domain_list = ""
        end
        local arr = util.split(domain_list, ',')
        local find = false
        for _, v in pairs(arr) do
            if v == domain then 
                find = true
                break
            end
        end
        --ngx.log(ngx.DEBUG, "hi domain_list: "..domain_list..','..#arr)
        if find == false then 
            table.insert(arr, host)
            -- ngx.log(ngx.DEBUG, "hi table: "..domain..','..#arr)
            ngx.shared.host_list:set('all', table.concat(arr, ','))
        end
    end
    -- ngx.header['Content-Type'] = 'text/html'
    -- ngx.say(res)
    return res
end
function _M:reloadsite()
    local domain = ngx.var.arg_domain
    local host = ngx.var.arg_host
    if domain ~= nil then _M:reloadDomain(domain) end
    if host ~= nil then _M:reloadHost(host) end
end
function _M:reloadDomain(domain)
    _M:load_server_from_file()
    
end
function _M:reloadHost(host)
    local host = ngx.var.arg_host
    _M:load_server_from_file()
    -- local cache_instance = require "cache" 
    -- cache.init()
    delCache("host", host)
end


function _M:tryload(hostname, force)
    local host = hostname
    local wild_domain = hostname
    ngx.log(ngx.DEBUG, "hostname: ".. hostname)
    if force == nil or not force then
    -- local name = _M.hprefix .. host
        local data = getCache("host", host)
        if data ~= nil then
            ngx.log(ngx.DEBUG, "data:"..data)
            if tostring(data) == "0" then
                ngx.log(ngx.DEBUG, "no domain")
                return nil
            end
            _M.configPath = host
            data = json.decode(data)
            return data
        end
        
        local arr = util.split(host, '.')
        local domain = "*"
        if #arr >= 2 then
            for i=2,#arr do
                domain = domain ..'.'.. arr[i]
            end
            wild_domain = domain
        end
        -- 二次判断是否存在泛解析的缓存，没有则等待httpapi，然后请求看看
        data = getCache("host", domain)
        if data ~= nil then
            ngx.log(ngx.DEBUG, "data:"..data)
            self.configPath = domain
            data = json.decode(data)
            return data
        end
    end
    
    local data = _M:getHostConfig(host, host, 'all')
    --ngx.log(ngx.INFO, "res:" .. data)
    if data == nil or data == "no" then 
        -- 说明站点不存在，可能是泛域名：
        -- 泛域名获取config
        local wild_cache = getCache("wild", wild_domain)
        if wild_cache ~= nil then
            ngx.log(ngx.DEBUG, "wild_cache:"..wild_cache)
        else
            ngx.log(ngx.DEBUG, "wild_cache is nil")
        end
        if wild_cache ~= nil and tostring(wild_cache) == "0" then
            -- pass
        else
            ngx.log(ngx.DEBUG, "wild_domain "..wild_domain.." hostname "..hostname)
            if wild_domain ~= hostname then -- 加载
                local data = _M:getHostConfig(wild_domain, wild_domain, 'all')
                ngx.log(ngx.INFO, "wild res:" ..wild_domain.. data)
                if data == nil or data == "no" then 
                    setCache("host", hostname, 0, 15)
                    setCache("wild", wild_domain, 0, 20)
                    return nil
                end
                data = util.string_split(data, "$#;#$")   -- domain$#:#$domaindata $#;#$ host
                if #data == 2 then
                    local domain = util.string_split(data[1], "$#:#$")
                    ngx.log(ngx.DEBUG, "domain"..data[1]..'#'..#domain)
                    if #domain == 2 then
                        if getCache("domain", domain[1]) == nil then
                            setCache("domain", domain[1], domain[2])
                        end
                    end
                    host = util.string_split(data[2], "$#:#$")
                    if #host == 2 then
                        self.configPath = host[1]
                        ngx.log(ngx.DEBUG, "setCache: "..host[1])
                        setCache("host", host[1], host[2])
                        return json.decode(host[2])
                    end
                end
                -- 试图获取api，如果不存在则泛域名不再请求
                setCache("wild", wild_domain, 0, 20)
            end
        end

        setCache("host", hostname, 0, 15)
        return nil
    end
    data = util.string_split(data, "$#;#$")   -- domain$#:#$domaindata $#;#$ host
    if #data == 2 then
        local domain = util.string_split(data[1], "$#:#$")
        ngx.log(ngx.DEBUG, "domain"..data[1]..'#'..#domain)
        if #domain == 2 then
            if getCache("domain", domain[1]) == nil then
                setCache("domain", domain[1], domain[2])
            end
        end
        host = util.string_split(data[2], "$#:#$")
        if #host == 2 then
            self.configPath = host[1]
            setCache("host", host[1], host[2])
            return json.decode(host[2])
        end
    end

    -- 域名不存在，记录缓存为0，下次查找直接返回
    setCache("host", hostname, 0, 15)
    --ngx.shared.domain:set(_M.hprefix .. host, 0, _M.configs['cache_time'] )
    return nil
end
function _M:load_ssl()
    -- local host = ngx.var.http_host
    if self.site_configs['ssl_enable'] then
        local data = getCache('ssl', self.site_configs['ssl'])
        if data then
            return util.split(data, "$#:#$")
        else
            data = _M:getHostConfig(self.site_configs['domain'], self.site_configs['ssl'], 'ssl')
            if data and data ~= "no" then
                setCache("ssl", self.site_configs['ssl'], data)
                return util.split(data, "$#:#$")
            else
                --local url = 'http://' .. host .. ngx.var.request_uri
                --ngx.redirect(url, ngx.HTTP_MOVED_PERMANENTLY)
            end
        end
    else
        --local url = 'http://' .. host .. ngx.var.request_uri
        --ngx.redirect(url, ngx.HTTP_MOVED_PERMANENTLY)
    end
end
function _M:load_cdn(hostname, force)
    local data
    if _M.configs == nil then
        _M.load_server_from_file()
        -- return nil
    end
    ngx.log(ngx.DEBUG, "hostname: "..hostname)
    local host = _M:tryload(hostname)
    -- ngx.log(ngx.INFO, "domain_id"..host['domain_id'])
    if host == nil or not host or host['domain'] == nil then
        if ngx.get_phase() == "ssl_cert" then
            return nil
        else
            if force ~= nil and force then
                host = _M:tryload(hostname)
            else
                _M:showdefault()
                return nil
            end
        end
    end
    if host == nil then
        return nil
    end
    if force == nil or not force then
        data = getCache("domain", host["domain"])
        if data ~= nil then
            data = require("json").decode(data)
            for k,v in pairs(host) do 
                --if k ~= 'domain' then
                data[k] = v 
                --end
            end -- domain host数据合并
            self.site_configs = data
            return self.site_configs
        end
    end
    -- ngx.say("no cache")
    local res = _M:getHostConfig(host['domain'], "", 'domain')-- util.http_get( _M.configs["data"] .. "?type=domain&domain=" .. host['domain'])
    if res ~= nil and res ~= "no" then
        local arr = util.string_split(res, "$#:#$")
        if #arr == 2 then
            data = json.decode(arr[2])
            setCache("domain", host["domain"], arr[2])
            for k,v in pairs(host) do 
                --if k ~= 'domain' then
                data[k] = v 
                -- end
            end -- domain host数据合并，从而覆盖主站点设置
            self.site_configs = data
        end
        return
    else
        if force ~= nil or not force then

            --pass
        else
            ngx.header['Content-Type'] = 'text/html'
            ngx.status = ngx.HTTP_FORBIDDEN
            ngx.say("no domain or data is not sync.")
            -- to cause quit the whole request rather than the current phase handler
            ngx.exit(ngx.HTTP_OK)
            return nil
        end
    end
    -- 从NFS读取
    --[[
    local file = io.open( _M.cprefix .. string.sub(domain, 1, 1) .. '/' .. domain .. '.json', "r")
    if file == nil then
       
    end
    ]]--
    if data ~= nil then
        ngx.shared.cdn_setting:set( self.prefix .. domain, data, _M.configs['cache_time'] )
        local cache_instance = initcache()
        cache_instance:set( self.prefix .. domain, data )
        cache_instance:close()
        self.site_configs = data
        return self.site_configs
    end
end

function _M:load_server_from_file()
    local config_dump_path = _M:home_path() .. "/configs/config.json"
    local file = io.open( config_dump_path, "r")
    
    if file == nil then
        return json.encode({["ret"]="error",["msg"]="config file not found"})
    end

    --file = io.open( "/tmp/config.json", "w");
    local data = file:read("*all");
    file:close();

    --save config hash in module
    local config_hash = ngx.md5(data)
    _M.config_hash = config_hash
    ngx.shared.status:set('vn_config_hash', config_hash )

    --ngx.log(ngx.STDERR, data)
    local tmp = dkjson.decode( data )
    if tmp ~= nil then
        --update config version if need
        -- local loop = true
        -- while loop do
        --     local handle = _M.version_updater[ tmp['config_version'] ] 
        --     if handle ~= nil then
        --         tmp = handle( tmp )
        --     else
        --         loop = false
        --     end 
        -- end

        if tmp['config_version'] ~= _M["configs"]["config_version"] then
            ngx.log(ngx.STDERR,"load config from config.json error,will use default config")
            ngx.log(ngx.STDERR,"Except Version:")
            ngx.log(ngx.STDERR, _M["configs"]["config_version"] )
            ngx.log(ngx.STDERR,"Config.json Version:")
            ngx.log(ngx.STDERR,tmp["config_version"])
        else
            _M["configs"] =  tmp
        end

        --return json.encode({["ret"]="success",['config']=_M["configs"]})
    else 
        ngx.log(ngx.STDERR,"config.json decode error")
        --return json.encode({["ret"]="error",["msg"]="config file decode error, will use default"})
    end
        
end 



function initcache()
    return cache.new(_M.configs['redis'], _M.configs['redis_port'], _M.configs['redis_pass'], _M.configs['redis_time'])
end

local prefix = "Site_"
local dprefix = "Domain_"
local hprefix = "Host_"
local sprefix = "SSL_"
local wprefix = "Wild_"
function getCache(types, name)
    local data
    if types == "wild" then
        name = wprefix .. name
        data = ngx.shared.domain:get(name)
    end
    if types == "domain" then
        name = dprefix .. name
        data = ngx.shared.domain:get(name)
    end
    if types == "host" then
        name = hprefix .. name
        data = ngx.shared.host:get(name)
    end
    if types == "ssl" then
        name = sprefix .. name
        data = ngx.shared.ssl:get(name)
    end
    if data ~= nil then 
        return data
    end
    -- data = cache.get(name)
    if data == nil then
        local cache_instance = initcache()
        data = (cache_instance:get(name))
        cache_instance:close()
        if data == ngx.null or data == "NULL" or not data then
            data = nil
        end
        return data
    end
    return nil
end
function setCache(types, name, value, times)
    if times == nil then
        times = 6 * 60
    end
    if types == "wild" then
        name = wprefix .. name
        ngx.shared.domain:set(name, value, times)
    end
    if types == "domain" then
        name = dprefix .. name
        ngx.shared.domain:set(name, value, times)
        
    end
    if types == "host" then
        name = hprefix .. name
        ngx.shared.host:set(name, value, times)
    end
    if types == "ssl" then
        name = sprefix .. name
        ngx.shared.ssl:set(name, value, times)
    end
    local cache_instance = initcache()
    cache_instance:set(name, value, times*1) -- 多存储一倍防止获取不到缓存和api
    cache_instance:close()
end
function delCache(types, name)
    if types == "wild" then
        name = wprefix .. name
        ngx.shared.domain:delete(name)
    end
    if types == "domain" then
        name = dprefix .. name
        ngx.shared.domain:delete(name)
    end
    if types == "host" then
        name = hprefix .. name
        ngx.shared.host:delete(name)
    end
    if types == "ssl" then
        name = sprefix .. name
        ngx.shared.ssl:delete(name)
    end
    local cache_instance = initcache()
    cache_instance:del(name)
    cache_instance:close()
end
return _M





